#include "gmMachineEx.h"

std::map<const gmType,gmClassDef*> gmMachineEx::ClassList;
std::multimap<const int,gmProxy*> gmMachineEx::CppFuncList;

std::multimap<const int,gmThread*> gmMachineEx::_ThreadList_AtCreate;
std::multimap<const int,gmThread*> gmMachineEx::_ThreadList_AtDestroy;

gmMachineEx::gmMachineEx( bool UseExtendedThreadFunctions )
{
	_GlobalFuncIndex = 0;

	if( UseExtendedThreadFunctions )
	{
		RegisterLibraryFunction( "threadEx", gmThreadEx );
		s_machineCallback = MachineCallback;
	}

	CreateUserType( "__unknown_type__" );
	CreateUserType( "__thread_launch_info__" );

	AT_CREATE = 0;
	AT_DESTROY = 1;
	AT_TIME = 2;
	m_global->Set( this, "AT_CREATE",	gmVariable( AllocUserObject( &AT_CREATE,	GM_THREAD_LAUNCH_INFO ) ) );
	m_global->Set( this, "AT_DESTROY",	gmVariable( AllocUserObject( &AT_DESTROY,	GM_THREAD_LAUNCH_INFO ) ) );
	m_global->Set( this, "AT_TIME",		gmVariable( AllocUserObject( &AT_TIME,		GM_THREAD_LAUNCH_INFO ) ) );

	gmClassDef* pDef = new gmClassDef;
	gmClassDef* pConstDef = new gmClassDef;

	if( pDef && pConstDef )
	{
		gmType Type = CreateUserType( "any_ptr" );
		gmType ConstType = CreateUserType( "any_ptr" );

		pDef->BaseClass = NULL;
		pConstDef->BaseClass = NULL;

		pDef->pTypeInfo = &typeid(void*);
		pConstDef->pTypeInfo = &typeid(const void*);

		ClassList.insert( std::make_pair( Type, pDef ) );
		ClassList.insert( std::make_pair( ConstType, pConstDef ) );
	}

	pDef = new gmClassDef;
	pConstDef = new gmClassDef;

	if( pDef && pConstDef )
	{
		gmType Type = CreateUserType( "c_string" );
		gmType ConstType = CreateUserType( "c_string" );

		pDef->BaseClass = NULL;
		pConstDef->BaseClass = NULL;

		pDef->pTypeInfo = &typeid(char*);
		pConstDef->pTypeInfo = &typeid(const char*);

		ClassList.insert( std::make_pair( Type, pDef ) );
		ClassList.insert( std::make_pair( ConstType, pConstDef ) );
	}

	_IndexList.reserve( 2048 );
};

bool gmMachineEx::OnCreateThread( const void* pContext )
{
	if( pContext )
	{
		const gmVariable* pFunc = ((gmThread*)pContext)->GetFunction();
		if( pFunc )
		{
			const int Ref = pFunc->m_value.m_ref;
			std::multimap<const int,gmThread*>::iterator first = gmMachineEx::_ThreadList_AtCreate.find(Ref);

			if( first != gmMachineEx::_ThreadList_AtCreate.end() )
			{
				std::multimap<const int,gmThread*>::const_iterator i = first;
				std::multimap<const int,gmThread*>::iterator last = gmMachineEx::_ThreadList_AtCreate.find(Ref+1);

				while( i != last )
				{
					if( i->second )
					{
						gmMachine* VM = i->second->GetMachine();
						if( VM )
							VM->Sys_SwitchState( i->second, gmThread::RUNNING );
					}
					
					++i;
				}

				gmMachineEx::_ThreadList_AtCreate.erase( first, last );
			}
		}
	}

	return false;
}

bool gmMachineEx::OnDestroyThread( const void* pContext )
{	
	if( pContext )
	{
		const gmVariable* pFunc = ((gmThread*)pContext)->GetFunction();
		if( pFunc )
		{
			const int Ref = pFunc->m_value.m_ref;
			std::multimap<const int,gmThread*>::iterator first = gmMachineEx::_ThreadList_AtDestroy.find(Ref);

			if( first != gmMachineEx::_ThreadList_AtDestroy.end() )
			{
				std::multimap<const int,gmThread*>::const_iterator i = first;
				std::multimap<const int,gmThread*>::iterator last = gmMachineEx::_ThreadList_AtDestroy.find(Ref+1);

				while( i != last )
				{
					if( i->second )
					{
						gmMachine* VM = i->second->GetMachine();
						if( VM )
							VM->Sys_SwitchState( i->second, gmThread::RUNNING );
					}
					
					++i;
				}

				gmMachineEx::_ThreadList_AtDestroy.erase( first, last );
			}
		}
	}

	return false;
}

bool gmMachineEx::MachineCallback( gmMachine*, gmMachineCommand Command, const void* pContext)	
{
	switch(Command)
	{
		case MC_THREAD_CREATE:
			return gmMachineEx::OnCreateThread( pContext );

		case MC_THREAD_DESTROY:
			return gmMachineEx::OnDestroyThread( pContext );
	}

	return false;
}


int gmMachineEx::gmThreadEx(gmThread* pThread)
{
	if( pThread )
	{
		gmMachine* VM = pThread->GetMachine();
		if( VM )
		{
			int NumParams = pThread->GetNumParams();
			int LaunchInfoIndex = NumParams-2;
			int LaunchParamIndex = NumParams-1;

			GM_ASSERT( NumParams >= 3 );
			GM_ASSERT( pThread->Param(0).m_type == GM_FUNCTION );
			GM_ASSERT( pThread->Param(LaunchInfoIndex).m_type == GM_THREAD_LAUNCH_INFO );

			if( NumParams >= 3 && pThread->Param(0).m_type == GM_FUNCTION && pThread->Param(LaunchInfoIndex).m_type == GM_THREAD_LAUNCH_INFO )
			{
				gmThread* NewThread = VM->CreateThread();
				if( NewThread )
				{
					const gmVariable* pThis = pThread->GetThis();
					if( !pThis )
						pThis = &gmVariable::s_null;

					gmFunctionObject* pFunc = (gmFunctionObject*)pThread->Param(0).m_value.m_ref;
					if( pFunc )
					{
						NewThread->Push( *pThis );
						NewThread->PushFunction( pFunc );

						for(int i = 1; i < LaunchInfoIndex; ++i)
							NewThread->Push(pThread->Param(i));
						NewThread->PushStackFrame( NumParams - 3 );

						gmUserObject* pGmObj = (gmUserObject*)pThread->Param(LaunchInfoIndex).m_value.m_ref;

						GM_ASSERT( pGmObj && pGmObj->m_user );
						if( pGmObj && pGmObj->m_user )
						{
							int LaunchInfo = *(int*)(pGmObj->m_user);

							if( LaunchInfo == 2 ) //AT_TIME
							{
								GM_ASSERT( pThread->Param(LaunchParamIndex).m_type == GM_INT || pThread->Param(LaunchParamIndex).m_type == GM_FLOAT );
								if( pThread->Param(LaunchParamIndex).m_type == GM_FLOAT )
									pThread->Param(LaunchParamIndex).m_value.m_int = (int) (pThread->Param(LaunchParamIndex).m_value.m_float * 1000.0f);
								else if( pThread->Param(LaunchParamIndex).m_type == GM_INT )
									pThread->Param(LaunchParamIndex).m_value.m_int *= 1000;

								NewThread->Sys_SetTimeStamp( VM->GetTime() + pThread->Param(LaunchParamIndex).m_value.m_int );
								VM->Sys_SwitchState( NewThread , gmThread::SLEEPING );
							}
							else
							{
								GM_ASSERT( pThread->Param(LaunchParamIndex).m_type == GM_FUNCTION );
								VM->Sys_SwitchState( NewThread, gmThread::BLOCKED );
								std::pair<const int,gmThread*> Pair = std::make_pair( pThread->Param(NumParams-1).m_value.m_ref, NewThread );

								if( LaunchInfo == 0) //AT_CREATE
									gmMachineEx::_ThreadList_AtCreate.insert( Pair );	
								else //AT_DESTROY
									gmMachineEx::_ThreadList_AtDestroy.insert( Pair );
							}

							return GM_OK;
						}
					}
				}
			}
		}
	}

	return GM_EXCEPTION;
}



gmMachineEx::~gmMachineEx()
{
	std::map<const gmType,gmClassDef*>::const_iterator i = ClassList.begin();
	std::multimap<const int,gmProxy*>::const_iterator j;
	std::map<const int,gmProxy*>::const_iterator k;

	gmClassDef* pDef = NULL;
	gmClassDef* pConstDef = NULL;

	while( i != ClassList.end() )
	{
		pDef = i->second;

		GM_ASSERT( pDef );
		if( pDef )
		{
			j = pDef->FuncList.begin();

			while( j != pDef->FuncList.end() )
			{
				if( j->second )
					delete j->second;

				++j;
			}

			pDef->FuncList.clear();

			k = pDef->VarList.begin();

			while( k != pDef->VarList.end() )
			{
				if( k->second )
					delete k->second;

				++k;
			}

			pDef->VarList.clear();
			delete pDef;
		}

		++i;
		
		pConstDef = i->second;
		if( pConstDef )
		{
			pConstDef->FuncList.clear(); //ce sont les meme proxy que pour les non const donc on n'y touche pas
			pConstDef->VarList.clear(); //pareil
			delete pConstDef;
		}

		++i;
	}

	ClassList.clear();

	j = CppFuncList.begin();

	while( j != CppFuncList.end() )
	{
		if( j->second )
			delete j->second;

		++j;
	}

	CppFuncList.clear();
	_GmFuncList.clear();

	std::vector<int*>::iterator l = _IndexList.begin();

	while( l != _IndexList.end() )
	{
		if( *l )
			delete *l;

		++l;
	}
	
	_IndexList.clear();
}

gmThread* gmMachineEx::LoadScript( const char* Filename )
{
	gmFileStream Script( Filename );
	unsigned int Size = Script.GetSize();
	char* memScript = NULL;

	if( Size <= 65536 )
	{
		if( Script.Read( _Script, 65536 ) != Size ) 
			return NULL;

		_Script[Script.Tell()]='\0';
		memScript = _Script;
	}
	else
	{
		memScript = new char[Size];

		if( !memScript || Script.Read( memScript, Size ) != Size ) 
			return NULL;

		memScript[Size]='\0';
	}

	int ThreadId = GM_INVALID_THREAD;
	ExecuteString( memScript, &ThreadId, false );

	if( Size > 65536 && memScript ) delete memScript;

	return GetThread( ThreadId );
}

gmThread* gmMachineEx::LoadLibrary( const char* Filename )
{
	gmFileStream Lib( Filename );

	int ThreadId= GM_INVALID_THREAD;
	ExecuteLib( Lib, &ThreadId, false );
	return GetThread( ThreadId );
}

bool gmMachineEx::CompileScript( const char* ScriptName, const char* LibraryName )
{		
	gmFileStream Script( ScriptName );
	unsigned int Size = Script.GetSize();
	char* memScript = NULL;

	if( Size <= 65536 )
	{
		if( Script.Read( _Script, 65536 ) != Size ) 
			return false;

		_Script[Script.Tell()]='\0';
		memScript = _Script;
	}
	else
	{
		memScript = new char[Size];

		if( !memScript || Script.Read( memScript, Size ) != Size ) 
			return false;

		memScript[Size]='\0';
	}

	gmFileStream Library( LibraryName, true );
	CompileStringToLib( memScript, Library );

	if( Size > 65536 && memScript ) delete memScript;

	return true;
}


unsigned int gmMachineEx::Hash( const char *Text )
{
	unsigned int iHash = 5381; 

	while( *Text != '\0' ) 
	{ 
		iHash = (iHash << 5) + iHash;
		iHash += *Text++; 
	} 

	return iHash;
}

gmThread* gmMachineEx::_BeginFuncCall( const char* FuncName, const gmVariable& This )
{
	if( FuncName )
	{
		unsigned int HashValue = Hash( FuncName );
		if( HashValue != _CurrentHash )
		{
			_CurrentHash = HashValue;
			std::map<const unsigned int, gmFunctionObject*>::const_iterator i = _GmFuncList.find( HashValue );
			
			if( i != _GmFuncList.end() )
				_CurrentFunc = i->second;
			else
			{
				gmVariable Temp( AllocStringObject(FuncName) );

				gmVariable gmFunc = m_global->Get(Temp);

				GM_ASSERT( gmFunc.m_type == GM_FUNCTION );
				if( gmFunc.m_type == GM_FUNCTION )
				{
					_CurrentFunc = (gmFunctionObject*)gmFunc.m_value.m_ref;

					if( _CurrentFunc ) 
						_GmFuncList.insert( std::make_pair( HashValue, _CurrentFunc ) );
				}
			}
		}

		if( _CurrentFunc )
		{
			gmThread* pThread = CreateThread();
			if( pThread )
			{
				pThread->Push(This);
				pThread->PushFunction( _CurrentFunc );

				return pThread;
			}
		}
	}

	return NULL;
}

gmVariableEx gmMachineEx::_EndFuncCall( gmThread* pThread, int ParamNumber )
{
	gmVariableEx ResultVar;

	if( pThread )
	{
		pThread->PushStackFrame( ParamNumber );
		GM_ASSERT( ParamNumber <= pThread->GetFunctionObject()->GetNumParams() );
		pThread->Sys_Execute(&ResultVar);
	}
		
	return ResultVar;
};

gmType gmMachineEx::Bind( bool isClass, const char* ClassOrNamespaceName, ...)
{	
	gmType Type = GM_NULL, ConstType = GM_NULL;
	gmClassDef* pDef = NULL, *pConstDef = NULL;
	gmTableObject* Table = NULL;
	std::vector<gmClassDef*> BaseClassList;
	
	if( !isClass && ClassOrNamespaceName )
	{
		Table = AllocTableObject();
		if( !Table ) return GM_NULL;

		m_global->Set( this, ClassOrNamespaceName, gmVariable( Table ) );
	}
	else
		Table = m_global;

	if( isClass )
	{
		GM_ASSERT( ClassOrNamespaceName );
		if(!ClassOrNamespaceName) return GM_NULL;

		Type = CreateUserType( ClassOrNamespaceName );
		ConstType = CreateUserType( ClassOrNamespaceName );

		m_types[Type].m_nativeOperators[O_EQ] = NULL;
		m_types[Type].m_nativeOperators[O_NEQ] = NULL;
		m_types[Type].m_nativeOperators[O_NOT] = NULL;

		m_types[ConstType].m_nativeOperators[O_EQ] = NULL;
		m_types[ConstType].m_nativeOperators[O_NEQ] = NULL;
		m_types[ConstType].m_nativeOperators[O_NOT] = NULL;

		pDef = new gmClassDef;
		pConstDef = new gmClassDef;

		if( !pDef || !pConstDef ) return GM_NULL;

		pDef->BaseClass = NULL;
		pConstDef->BaseClass = NULL;
	}

	va_list pArg;
	va_start( pArg, ClassOrNamespaceName );

	gmVariable TempVar;
	int LocalFuncIndex = 0;
	std::pair<int,gmProxy*> Pair;
	gmFunctionObject* pFuncObj = NULL;
	gmProxy* pProxy = NULL;
	int* pIndex = NULL;
	const char* Name = NULL;
	bool End = false;

	while( !End )
	{
		pProxy = va_arg( pArg, gmProxy* );

		if( pProxy )
		{
			switch( pProxy->GetProxyType() )
			{
				case FUNCTION_TYPE:
					{
						Name = va_arg( pArg, const char* );
						if( Name )
						{
							TempVar = Table->Get( this, Name );

							if( TempVar.m_type != GM_FUNCTION )
							{
								pIndex = new int(_GlobalFuncIndex);
								if( pIndex )
								{
									_IndexList.push_back( pIndex );

									if( isClass )
										RegisterLibraryFunction( Name, &FunctionDispatcher, NULL, pIndex );
									else
										RegisterLibraryFunction( Name, &FunctionDispatcher, ClassOrNamespaceName, pIndex );

									CppFuncList.insert( std::make_pair( _GlobalFuncIndex++, pProxy ) );
								}
							}
							else
							{
								pFuncObj = ((gmFunctionObject*)TempVar.m_value.m_ref);
								if( pFuncObj && pFuncObj->m_cUserData )
									CppFuncList.insert( std::make_pair( *(int*)pFuncObj->m_cUserData, pProxy ) );
							}
						}
					} break;

				case VARIABLE_TYPE:
					{
						GM_ASSERT( isClass );
						if( isClass )
						{
							Name = va_arg( pArg, const char* );
							if( Name )
							{
								Pair = std::make_pair( Hash(Name), pProxy );
								pDef->VarList.insert( Pair );
								pConstDef->VarList.insert( Pair );
							}
						}
					} break;

				case METHODE_TYPE:
					{
						GM_ASSERT( isClass );
						if( isClass )
						{
							Name = va_arg( pArg, const char* );
							if( Name )
							{
								TempVar = GetTypeVariable( Type, gmVariable( AllocStringObject( Name ) ) );

								if( TempVar.m_type != GM_FUNCTION )
								{
									pIndex = new int(LocalFuncIndex);
									if( pIndex )
									{
										_IndexList.push_back( pIndex );
										RegisterTypeLibraryFunction( Type, Name, &MethodeDispatcher, pIndex );
										Pair = std::make_pair( LocalFuncIndex++, pProxy ) ;

										pDef->FuncList.insert( Pair );

										if( pProxy->isConst() )
										{
											RegisterTypeLibraryFunction( ConstType, Name, &MethodeDispatcher, pIndex );
											pConstDef->FuncList.insert( Pair );
										}
									}
								}
								else
								{
									pFuncObj = ((gmFunctionObject*)TempVar.m_value.m_ref);
									if( pFuncObj && pFuncObj->m_cUserData )
									{
										Pair = std::make_pair( *(int*)pFuncObj->m_cUserData, pProxy );
										pDef->FuncList.insert( Pair );

										if( pProxy->isConst() )
											pConstDef->FuncList.insert( Pair );
									}
								}
							}
						}

					} break;

				case CONSTRUCTOR_TYPE:
					{
						GM_ASSERT( isClass );
						if( isClass )
						{
							TempVar = m_global->Get( this, ClassOrNamespaceName );

							if( TempVar.m_type != GM_FUNCTION )
							{
								pIndex = new int(_GlobalFuncIndex);
								if( pIndex )
								{
									_IndexList.push_back( pIndex );
									RegisterLibraryFunction( ClassOrNamespaceName, &FunctionDispatcher, NULL, pIndex );
									CppFuncList.insert( std::make_pair( _GlobalFuncIndex++, pProxy ) );
								}
							}
							else
							{
								pFuncObj = ((gmFunctionObject*)TempVar.m_value.m_ref);
								if( pFuncObj && pFuncObj->m_cUserData )
									CppFuncList.insert( std::make_pair( *(int*)pFuncObj->m_cUserData, pProxy ) );
							}
						}

					} break;

				case FUNCTION_OPERATOR_TYPE:
					{
						GM_ASSERT( isClass );
						if( isClass )
						{
							gmOperator Op = va_arg( pArg, gmOperator );

							if( Op != O_MAXOPERATORS )
							{
								pFuncObj = GetTypeOperator( Type, Op );
								if( pFuncObj && pFuncObj->m_cUserData )
								{
									Pair = std::make_pair( *(int*)pFuncObj->m_cUserData, pProxy );
									pDef->FuncList.insert( Pair );

									if( pProxy->isConst() )
										pConstDef->FuncList.insert( Pair );
								}
								else
								{
									pIndex = new int(LocalFuncIndex);
									if( pIndex )
									{
										_IndexList.push_back( pIndex );
										pFuncObj = AllocFunctionObject( &OperatorDispatcher );
										if( pFuncObj )
										{
											pFuncObj->m_cUserData = pIndex;
											RegisterTypeOperator( Type, Op, pFuncObj );
											RegisterTypeOperator( ConstType, Op, pFuncObj ); //hmm pas sur...comment savoir s'il ne modifie pas l'objet?...il faudrait tester les parametres pour voir si l'objet est pris en tant que const ref/ptr ou pas
											pDef->FuncList.insert( std::make_pair( LocalFuncIndex++, pProxy ) );
										}
									}
								}
							}
						}

					} break;

				case METHODE_OPERATOR_TYPE:
					{
						GM_ASSERT( isClass );
						if( isClass )
						{
							gmOperator Op = va_arg( pArg, gmOperator );

							if( Op != O_MAXOPERATORS )
							{
								pFuncObj = GetTypeOperator( Type, Op );
								if( pFuncObj && pFuncObj->m_cUserData )
								{
									Pair = std::make_pair( *(int*)pFuncObj->m_cUserData, pProxy );
									pDef->FuncList.insert( Pair );

									if( pProxy->isConst() )
										pConstDef->FuncList.insert( Pair );
								}
								else
								{
									pIndex = new int(LocalFuncIndex);
									if( pIndex )
									{
										_IndexList.push_back( pIndex );
										pFuncObj = AllocFunctionObject( &OperatorDispatcher );
										if( pFuncObj )
										{
											pFuncObj->m_cUserData = pIndex;
											RegisterTypeOperator( Type, Op, pFuncObj );
											
											if( Op == O_GETIND )
											{
												RegisterTypeOperator( Type, O_SETIND, pFuncObj );

												if( pProxy->isConst() )
													RegisterTypeOperator( ConstType, O_SETIND, pFuncObj );
											}

											Pair = std::make_pair( LocalFuncIndex++, pProxy );
											pDef->FuncList.insert( Pair );
											
											if( pProxy->isConst() )
											{
												RegisterTypeOperator( ConstType, Op, pFuncObj );
												pConstDef->FuncList.insert( Pair );
											}
										}
									}
								}
							}
						}

					} break;
			}
		}
		else
		{
			int DataType = va_arg( pArg, int );
			switch( DataType )
			{
			case DESTRUCTOR_VIRTUAL_TYPE:
				{	
					GM_ASSERT( isClass );
					if( isClass )
					{
						gmGCDestructCallback pDestructor = va_arg( pArg, gmGCDestructCallback );
						if( pDestructor )
						{
							RegisterUserCallbacks( Type, NULL, pDestructor );
							RegisterUserCallbacks( ConstType, NULL, pDestructor );
						}
					}

				} break;

			case BASECLASS_VIRTUAL_TYPE:
				{	
					GM_ASSERT( isClass );
					if( isClass )
					{
						type_info* pInfo = va_arg( pArg, type_info* );
						if( pInfo )
						{
							gmType BaseType = GetTypeId( pInfo );
							std::map<const gmType,gmClassDef*>::const_iterator i = ClassList.find( BaseType );
							if( i != ClassList.end() )
							{
								BaseClassList.push_back( i->second );
								//pDef->BaseClass = i->second;
								++i;
								BaseClassList.push_back( i->second );
								//pConstDef->BaseClass = i->second;
							}
						}
					}

				} break;

			case POOL_VIRTUAL_TYPE: //code incertain pour l'instant
				{
					GM_ASSERT( isClass );
					if( isClass )
					{
						pProxy = va_arg( pArg, gmProxy* );
					
					//	GM_ASSERT( pProxy );
						TempVar = m_global->Get( this, ClassOrNamespaceName  );

						if( TempVar.m_type != GM_FUNCTION )
						{
							pIndex = new int(_GlobalFuncIndex);
							_IndexList.push_back( pIndex );
							RegisterLibraryFunction( ClassOrNamespaceName, &FunctionDispatcher, NULL, pIndex );
							CppFuncList.insert( std::make_pair( _GlobalFuncIndex++, pProxy ) );
						}
						else
						{
							pFuncObj = ((gmFunctionObject*)TempVar.m_value.m_ref);
							CppFuncList.insert( std::make_pair( *(int*)pFuncObj->m_cUserData, pProxy ) );
						}

						pProxy = va_arg( pArg, gmProxy* );
						pDef->FuncList.insert( std::make_pair( 4294967295, pProxy ) );
						pConstDef->FuncList.insert( std::make_pair( 4294967295, pProxy ) );

						gmGCDestructCallback pDestructor = va_arg( pArg, gmGCDestructCallback );
						RegisterUserCallbacks( Type, NULL, pDestructor );
						RegisterUserCallbacks( ConstType, NULL, pDestructor );
					}

				} break;

			case END_VIRTUAL_TYPE:
				{ 
					End = true;
					if( isClass )
					{
						pDef->pTypeInfo = va_arg( pArg, type_info* );
						pConstDef->pTypeInfo = va_arg( pArg, type_info* );
					}

				} break;
			}
		}
	}

	va_end( pArg );

	if( isClass )
	{
		RegisterTypeOperator( Type, O_GETDOT, NULL, &gmMachineEx::OpGetDot );
		RegisterTypeOperator( Type, O_SETDOT, NULL, &gmMachineEx::OpSetDot );
		RegisterTypeOperator( ConstType, O_GETDOT, NULL, &gmMachineEx::OpGetDot );

		pDef->BaseClass = new const gmClassDef*[ BaseClassList.size() / 2 ];
		pConstDef->BaseClass = new const gmClassDef*[ BaseClassList.size() / 2 ];

		if( pDef->BaseClass && pConstDef->BaseClass )
		{
			std::vector<gmClassDef*>::const_iterator t = BaseClassList.begin();
			int nx = 0;

			while( t != BaseClassList.end() )
			{
				pDef->BaseClass[nx] = *t;
				++t;
				pConstDef->BaseClass[nx] = *t;

				++nx;
			}

			pDef->NumBaseClass = nx;
			pConstDef->NumBaseClass = nx;
		}
	/*
		if( pDef->BaseClass )
		{
			gmType BaseType = GetTypeId( pDef->BaseClass->pTypeInfo );
			gmTableObject* pBaseTable = m_types[BaseType].m_variables;
			gmTableNode* pBaseNode = NULL;
			gmTableObject* pThisTable = m_types[Type].m_variables;
			gmTableNode* pThisNode = NULL;

			gmTableIterator it = -2;

			while( (pBaseNode = pBaseTable->GetNext(it) ) != NULL )
			{
				if( pBaseNode->m_value.m_type == GM_FUNCTION )
					RegisterTypeLibraryFunction( Type, ((gmStringObject*)pBaseNode->m_key.m_ref)->GetString(), &MethodeDispatcher, pIndex );
			}

			for( gmOperator Op = O_GETIND; Op != O_MAXOPERATORS; ++Op )
			{
				if( GetTypeOperator( Type, Op ) == NULL ) //plus GetTypeOperator( ConstType, Op ) ?
				{
					pFuncObj = GetTypeOperator( BaseType, Op ); //plus GetTypeOperator( BaseType+1, Op ) pour const?
					if( pFuncObj )
					{
						RegisterTypeOperator( Type, Op, pFuncObj ); //prob car l'index dans pFuncObj corrrespond a celui du PARENT!
						RegisterTypeOperator( ConstType, Op, pFuncObj ); //if is const
					}
				}
			}
		}*/

		ClassList.insert( std::make_pair( Type, pDef ) );
		ClassList.insert( std::make_pair( ConstType, pConstDef ) );

		std::map<const gmType,gmClassDef*>::const_iterator i = ClassList.begin();
		std::multimap<const int,gmProxy*>::const_iterator j;
		std::map<const int,gmProxy*>::const_iterator k;

		while( i != ClassList.end() )
		{
			if( i->second )
			{
				j = i->second->FuncList.begin();
				while( j != i->second->FuncList.end() )
				{
					if( j->second )
						j->second->UpdateTypes();

					++j;
				}

				k = i->second->VarList.begin();

				while( k != i->second->VarList.end() )
				{
					if( k->second )
						k->second->UpdateTypes();

					++k;
				}
			}

			++i; ++i; //passer le second type qui est const et donc a les meme proxy
		}

		std::multimap<const int,gmProxy*>::const_iterator l = CppFuncList.begin();

		while( l != CppFuncList.end() )
		{
			if( l->second )
				l->second->UpdateTypes();

			++l;
		}
	}

	return Type;
}




int GM_CDECL gmMachineEx::FunctionDispatcher( gmThread* pThread )
{
	if( pThread )
	{
		const gmFunctionObject* pFunc = pThread->GetFunctionObject();
		if( pFunc && pFunc->m_cUserData )
		{
			int* pIndex = (int*)pFunc->m_cUserData;
			if( pIndex )
			{
				std::multimap<const int,gmProxy*>::const_iterator i = gmMachineEx::CppFuncList.find( *pIndex );
				std::multimap<const int,gmProxy*>::const_iterator end = gmMachineEx::CppFuncList.find( *pIndex + 1 );

				gmProxy* pProxy = NULL, *TempProxy = NULL;
				int NumParam = pThread->GetNumParams();
				int ParamCoeff = NumParam * 3;
				int TotalCoeff, BestCoeff = 0, TempCoeff;

				#ifdef _DEBUG
					bool Ambiguity = false;
				#endif

				while( i != end )
				{
					TotalCoeff = 0;
					TempProxy = i->second;

					if( TempProxy )
					{
						for( int x = 0; x < NumParam; ++x )
						{
							TempCoeff = TempProxy->CompareTypes( pThread->Param(x).m_type, x );

							if( TempCoeff )
								TotalCoeff += TempCoeff;
							else
							{
								TotalCoeff = -1;
								break;
							}
						}

						if( TotalCoeff == ParamCoeff )
						{
							pProxy = TempProxy;

							#ifdef _DEBUG
								Ambiguity = false;
							#endif

							break;
						}
						else if( TotalCoeff > BestCoeff )
						{
							BestCoeff = TotalCoeff;
							pProxy = TempProxy;

							#ifdef _DEBUG
								Ambiguity = false;
							#endif
						}
						#ifdef _DEBUG
						else if( TotalCoeff == BestCoeff )
							Ambiguity = true;
						#endif
					}

					++i;
				}

				if( pProxy )
				{
					#ifdef _DEBUG
						if( Ambiguity )
							pThread->GetMachine()->GetLog().LogEntry( "Ambiguous call to an overloaded function." );
					#endif

					pThread->Push( (*pProxy)(pThread, NULL ) );
					return GM_OK;
				}
			}
		}
	}

	return GM_EXCEPTION;
}

int GM_CDECL gmMachineEx::MethodeDispatcher( gmThread* pThread )
{
	if( pThread )
	{
		const gmVariable* pThis = pThread->GetThis();
		if( pThis )
		{
			gmUserObject* pGmObj = (gmUserObject*)pThis->m_value.m_ref;
			if( pGmObj )
			{
				const gmFunctionObject* pFunc = pThread->GetFunctionObject();
				if( pFunc && pFunc->m_cUserData )
				{
					int* pIndex = (int*)pFunc->m_cUserData;
					if( pIndex )
					{
						std::map<const gmType,gmClassDef*>::const_iterator j = gmMachineEx::ClassList.find( pGmObj->m_userType );

						if( j != gmMachineEx::ClassList.end() )
						{
							const gmClassDef* pDef = NULL;

							if( *pIndex >= 0 )
								pDef = j->second;
							else
							{
								*pIndex = -(*pIndex);
/*								if( j->second )
									pDef = j->second->BaseClass;*/
							}

							if( pDef )
							{
								std::multimap<const int,gmProxy*>::const_iterator i = pDef->FuncList.find( *pIndex );
								std::multimap<const int,gmProxy*>::const_iterator end = pDef->FuncList.find( *pIndex + 1 );

								gmProxy* pProxy = NULL, *TempProxy = NULL;
								int NumParam = pThread->GetNumParams();
								int ParamCoeff = NumParam * 3;
								int TotalCoeff, BestCoeff = 0, TempCoeff;

								#ifdef _DEBUG
									bool Ambiguity = false;
								#endif

								while( i != end )
								{
									TotalCoeff = 0;
									TempProxy = i->second;

									for( int x = 0; x < NumParam; ++x )
									{
										TempCoeff = TempProxy->CompareTypes( pThread->Param(x).m_type, x );

										if( TempCoeff )
											TotalCoeff += TempCoeff;
										else
										{
											TotalCoeff = -1;
											break;
										}
									}

									if( TotalCoeff == ParamCoeff )
									{
										pProxy = TempProxy;

										#ifdef _DEBUG
											Ambiguity = false;
										#endif

										break;
									}
									else if( TotalCoeff > BestCoeff )
									{
										BestCoeff = TotalCoeff;
										pProxy = TempProxy;

										#ifdef _DEBUG
											Ambiguity = false;
										#endif
									}
									#ifdef _DEBUG
									else if( TotalCoeff == BestCoeff )
										Ambiguity = true;
									#endif

									++i;
								}

								if( pProxy )
								{
									#ifdef _DEBUG
										if( Ambiguity )
											pThread->GetMachine()->GetLog().LogEntry( "Ambiguous call to an overloaded methode." );
									#endif

									pThread->Push( (*pProxy)(pThread, pGmObj->m_user ) );
									return GM_OK;
								}
							}
						}
					}
				}
			}
		}
	}

	return GM_EXCEPTION;
}


int GM_CDECL gmMachineEx::OperatorDispatcher( gmThread* pThread )
{
	if( pThread )
	{
		const gmFunctionObject* pFunc = pThread->GetFunctionObject();
		if( pFunc && pFunc->m_cUserData )
		{
			int* pIndex = (int*)pFunc->m_cUserData;
			if( pIndex )
			{
				gmUserObject* pGmObj = NULL;

				if( pThread->Param(0).m_type > GM_USER )
					pGmObj = (gmUserObject*)pThread->Param(0).m_value.m_ref;
				else if( pThread->GetNumParams() > 1 && pThread->Param(1).m_type > GM_USER )
					pGmObj = (gmUserObject*)pThread->Param(1).m_value.m_ref;

				if( pGmObj )
				{
					std::map<const gmType,gmClassDef*>::iterator j = gmMachineEx::ClassList.find( pGmObj->m_userType );

					if( j != gmMachineEx::ClassList.end() )
					{
						const gmClassDef* pDef = NULL;

						if( *pIndex >= 0 )
							pDef = j->second;
						else
						{
							*pIndex = -(*pIndex);
/*							if( j->second )
								pDef = j->second->BaseClass;*/
						}

						if( pDef )
						{
							std::multimap<const int,gmProxy*>::const_iterator i = pDef->FuncList.find( *pIndex );
							std::multimap<const int,gmProxy*>::const_iterator end = pDef->FuncList.find( *pIndex + 1 );

							gmProxy* pProxy = NULL, *TempProxy = NULL;
							int TotalCoeff, BestCoeff = 0, TempCoeff;
							int NumParam;

							#ifdef _DEBUG
								bool Ambiguity = false;
							#endif

							while( i != end )
							{
								NumParam = 0;
								TotalCoeff = 0;
								TempProxy = i->second;

								if( TempProxy )
								{
									if( TempProxy->GetProxyType() == FUNCTION_OPERATOR_TYPE )
									{
										NumParam = pThread->GetNumParams();

										for( int x = 0; x < NumParam; ++x )
										{
											TempCoeff = TempProxy->CompareTypes( pThread->Param(x).m_type, x );

											if( TempCoeff )
												TotalCoeff += TempCoeff;
											else
											{
												TotalCoeff = 0;
												break;
											}
										}
									}
									else
									{
										if( pThread->GetNumParams() > 1 )
											NumParam = 1;

										if( pThread->Param(0).m_type > GM_USER )
										{
											if( NumParam == 1 )
											{
												TotalCoeff = TempProxy->CompareTypes( pThread->Param(1).m_type, 0 );
												if( TotalCoeff == 0 )
													break;
											}
										}
									}
									
									if( TotalCoeff == NumParam * 3 )
									{
										pProxy = TempProxy;

										#ifdef _DEBUG
											Ambiguity = false;
										#endif

										break;
									}
									else if( TotalCoeff > BestCoeff )
									{
										BestCoeff = TotalCoeff;
										pProxy = TempProxy;

										#ifdef _DEBUG
											Ambiguity = false;
										#endif
									}
									#ifdef _DEBUG
									else if( TotalCoeff == BestCoeff )
										Ambiguity = true;
									#endif
								}

								++i;
							}

							if( pProxy )
							{
								#ifdef _DEBUG
									if( Ambiguity )
									{
										if( pProxy->GetProxyType() == FUNCTION_OPERATOR_TYPE )
											pThread->GetMachine()->GetLog().LogEntry( "Ambiguous call to an overloaded non-member operator." );
										else
											pThread->GetMachine()->GetLog().LogEntry( "Ambiguous call to an overloaded member operator." );
									}
								#endif

								pThread->Push( (*pProxy)(pThread, pGmObj->m_user ) );//NULL?
								return GM_OK;
							}
						}
					}
				}	
			}
		}
	}

	return GM_EXCEPTION;
}

template<> int ATC<int>( const gmVariable* pVar )
{
	if( pVar )
	{
		switch( pVar->m_type )
		{
		case GM_INT:
			return pVar->m_value.m_int;

		case GM_FLOAT:
			return (int)pVar->m_value.m_float;

		case GM_C_STRING:
		case GM_CONST_C_STRING:
			{
				const char* String = (const char*)pVar->m_value.m_ref;
				if( String )
					return atoi( String );

			} break;

		case GM_STRING:
			{
				gmStringObject* pStringObj = (gmStringObject*)pVar->m_value.m_ref;
				if( pStringObj )
				{
					const char* String = pStringObj->GetString();
					if( String )
						return atoi( String );
				}
			}
		}
	}
	
	return 0;
}

template<> float ATC<float>( const gmVariable* pVar )
{
	if( pVar )
	{
		switch( pVar->m_type )
		{
		case GM_INT:
			return (float)pVar->m_value.m_int;

		case GM_FLOAT:
			return pVar->m_value.m_float;

		case GM_C_STRING:
		case GM_CONST_C_STRING:
			{
				const char* String = (const char*)pVar->m_value.m_ref;
				if( String )
					return (float)atof( String );

			} break;

		case GM_STRING:
			{
				gmStringObject* pStringObj = (gmStringObject*)pVar->m_value.m_ref;
				if( pStringObj )
				{
					const char* String = pStringObj->GetString();
					if( String )
						return (float)atof( String );
				}
			}
		}
	}

	return 0.0f;
}

template<> const char* ATC<const char*>( const gmVariable* pVar )
{
	if( pVar )
	{
		switch( pVar->m_type )
		{
		case GM_INT:
			{
				static char Buffer[64];
				sprintf( Buffer, "%d", pVar->m_value.m_int );
				return Buffer;
			}

		case GM_FLOAT:
			{
				static char Buffer[64];
				sprintf( Buffer, "%f", pVar->m_value.m_float );
				return Buffer;
			}

		case GM_C_STRING:
		case GM_CONST_C_STRING:
			{
				const char* String = (const char*)pVar->m_value.m_ref;
				if( String )
					return String;

			} break;

		case GM_STRING:
			{
				gmStringObject* pStringObj = (gmStringObject*)pVar->m_value.m_ref;
				if( pStringObj )
				{
					const char* String = pStringObj->GetString();
					if( String )
						return String;
				}
			}
		}
	}
	
	return "";
}



void GM_CDECL gmMachineEx::OpGetDot(gmThread* pThread, gmVariable* Operand)
{	
	std::map<const gmType,gmClassDef*>::const_iterator i = gmMachineEx::ClassList.find(Operand->m_type);

	GM_ASSERT( i != gmMachineEx::ClassList.end() );
	if( i != gmMachineEx::ClassList.end() )
	{
		gmUserObject* pGmObj = (gmUserObject*)Operand->m_value.m_ref;
		if( pGmObj )
		{
			void* pObj = pGmObj->m_user;
			if( pObj )
			{
				gmStringObject* pStringObj = (gmStringObject*)Operand[1].m_value.m_ref;
				if( pStringObj )
				{
					const char* String = pStringObj->GetString();
					if( String )
					{
						unsigned int HashValue = gmMachineEx::Hash( String );

						if( i->second )
						{
							std::map<const int,gmProxy*>::const_iterator j = i->second->VarList.find( HashValue );

							gmProxy* pProxy = NULL;

							if( j == i->second->VarList.end() )
							{
/*								const gmClassDef* BaseClass = i->second->BaseClass;

								if( BaseClass )
								{
									j = BaseClass->VarList.find( HashValue );
									if( j != BaseClass->VarList.end() )
										pProxy = j->second;
								}*/
							}
							else
								pProxy = j->second;

							if( pProxy )
							{
								*Operand = (*pProxy)( pThread, pObj );
								return;	
							}
						}
					}
				}
			}
		}
	}

	Operand->Nullify();
}


void GM_CDECL gmMachineEx::OpSetDot(gmThread* pThread, gmVariable* Operand) //checker si Obj.Var est de type const?
{
	std::map<const gmType,gmClassDef*>::const_iterator i = gmMachineEx::ClassList.find(Operand->m_type);

	if( i != gmMachineEx::ClassList.end() )
	{
		gmUserObject* pGmObj = (gmUserObject*)Operand->m_value.m_ref;

		if( pGmObj )
		{
			void* pObj = pGmObj->m_user;
			if( pObj )
			{
				gmStringObject* pStringObj = (gmStringObject*)Operand[1].m_value.m_ref;
				if( pStringObj )
				{
					const char* String = pStringObj->GetString();
					if( String )
					{
						unsigned int HashValue = gmMachineEx::Hash( ((gmStringObject*)(Operand+2)->m_value.m_ref)->GetString() );

						std::map<const int,gmProxy*>::const_iterator j = i->second->VarList.find( HashValue );

						gmProxy* pProxy = NULL;

						if( j == i->second->VarList.end() )
						{
/*							const gmClassDef* BaseClass = i->second->BaseClass;

							if( BaseClass )
							{
								j = BaseClass->VarList.find( HashValue );
								if( j != BaseClass->VarList.end() )
									pProxy = j->second;
							}*/
						}
						else
							pProxy = j->second;

						if( pProxy )
						{ //if( ! pProxy->isConst() )
							if( pProxy->CompareTypes( Operand[1].m_type, 0 ) )
							{
								(*pProxy)( pObj, Operand[1] );
								return;	
							}
						}
					}
				}
			}
		}
	}

	Operand->Nullify();
}




gmOperator gmMachineEx::CppOpToGm( const char* Op, bool isUnary )
{
	if( Op )
	{
		if( !strcmp( Op, "+" ) )
		{
			if( isUnary )
				return O_POS;
			else
				return O_ADD;
		}
		else if( !strcmp( Op, "-" ) )
		{
			if( isUnary )
				return O_NEG;
			else
				return O_SUB;
		}
		else if( !strcmp( Op, "*" ) )
			return O_MUL;
		else if( !strcmp( Op, "/" ) )
			return O_DIV;
		else if( !strcmp( Op, "%" ) )
			return O_REM;
		else if( !strcmp( Op, "|" ) )
			return O_BIT_OR;
		else if( !strcmp( Op, "^" ) )
			return O_BIT_XOR;
		else if( !strcmp( Op, "&" ) )
			return O_BIT_AND;
		else if( !strcmp( Op, "<<" ) )
			return O_BIT_SHIFTLEFT;
		else if( !strcmp( Op, ">>" ) )
			return O_BIT_SHIFTRIGHT;
		else if( !strcmp( Op, "~" ) )
			return O_BIT_INV;
		else if( !strcmp( Op, "<" ) )
			return O_LT;
		else if( !strcmp( Op, ">" ) )
			return O_GT;
		else if( !strcmp( Op, "<=" ) )
			return O_LTE;
		else if( !strcmp( Op, ">=" ) )
			return O_GTE;
		else if( !strcmp( Op, "==" ) )
			return O_EQ;
		else if( !strcmp( Op, "!=" ) )
			return O_NEQ;
		else if( !strcmp( Op, "!" ) )
			return O_NOT;
		else if( !strcmp( Op, "[]" ) )
			return O_GETIND;
	}

	return O_MAXOPERATORS;
}


gmType GetTypeId( type_info* pInfo )
{
	if( pInfo )
	{
		std::map<const gmType,gmClassDef*>::const_iterator i = gmMachineEx::ClassList.begin();

		while( i != gmMachineEx::ClassList.end() )
		{
			if( i->second )
			{
				const type_info* pClassInfo = i->second->pTypeInfo;
				if( pClassInfo )
				{
					if( *pClassInfo == *pInfo )
						return i->first;
				}
			}
			else
				++i;
		}
	}

	return GM_UNKNOWN;
};