#ifndef GM_MACHINE_EX_H
#define GM_MACHINE_EX_H

#include <map>
#include <vector>
#include <typeinfo>

#include "gmThread.h"

template<typename TYPE> TYPE ATC( const gmVariable* );
template<> int ATC<int>( const gmVariable* );
template<> float ATC<float>( const gmVariable* );
template<> const char* ATC<const char*>( const gmVariable* );

#include "gmTmp.h"
#include "gmVariableEx.h"
#include "gmFileStream.h"

class gmProxy
{
public:
	virtual ~gmProxy(){}
	virtual int GetProxyType() const = 0;
	virtual bool isConst() const { return false; }
	virtual gmVariable operator()(gmThread*, void*) const = 0;
	virtual void operator()(void*, gmVariable&) const {}
	virtual int CompareTypes( gmType, int ) const = 0;
	virtual void UpdateTypes() = 0;
};

#include "gmVariableProxy.h"
#include "gmFunctionProxy.h"

struct gmClassDef
{
	std::multimap<const int,gmProxy*> FuncList;
	std::map<const int,gmProxy*> VarList;
	const type_info* pTypeInfo;
	const gmClassDef** BaseClass;
	int NumBaseClass;
};

#define CLASS_NAME( Name ) true, Name
#define NAMESPACE( Name ) false, Name
#define GLOBAL_NAMESPACE false, NULL
#define FUNCTION( Func, ... ) CreateFunctionProxy<__VA_ARGS__>( &Func ), #Func
#define METHODE( Meth, ... )  CreateMethodeProxy<__VA_ARGS__>( &CLASS::Meth ), #Meth
#define VARIABLE( Var, ... ) CreateVariableProxy( &CLASS::Var, __VA_ARGS__ ), #Var
#define CONSTRUCTOR( ... ) CreateConstructorProxy<CLASS,__VA_ARGS__>()
#define DESTRUCTOR()  NULL, DESTRUCTOR_VIRTUAL_TYPE, (void(*)(gmMachine*,gmUserObject*))&gmDestructor<true,CLASS>
#define OPERATOR(Op,...) CreateOperatorProxy<__VA_ARGS__>( &CLASS::operator##Op ), gmMachineEx::CppOpToGm(#Op, isUnary<void,__VA_ARGS__>(&CLASS::operator##Op) )
#define NM_OPERATOR(Op,...) CreateOperatorProxy<__VA_ARGS__>( &operator##Op ), gmMachineEx::CppOpToGm(#Op, isUnary<void,__VA_ARGS__>(&operator##Op) )
#define BASE_CLASS(Class) NULL, BASECLASS_VIRTUAL_TYPE, &typeid(Class*)
#define END NULL, END_VIRTUAL_TYPE, &typeid(CLASS*), &typeid(const CLASS*)
#define POOL( Get, Free ) NULL, POOL_VIRTUAL_TYPE, CreateFunctionProxy( &Get ), CreateFunctionProxy( &Free ), (void(*)(gmMachine*,gmUserObject*))&gmDestructor<false,CLASS>

template<bool isAutoDestructed, class _CLASS_> void gmDestructor( gmMachine* VM, gmUserObject* pGmObj )
{
	if( pGmObj && pGmObj->m_user )
		IF<isAutoDestructed>::Destruct<_CLASS_>( pGmObj, VM );
}

class gmMachineEx : public gmMachine
{
public:
	gmMachineEx( bool UseExtendedThreadFunctions = false );
	~gmMachineEx();

	gmThread* LoadScript( const char* Filename );
	gmThread* LoadLibrary( const char* Filename );

	inline gmThread::State ExecuteScript( const char* Filename ) { gmThread* pThread = LoadScript( Filename ); return pThread->Sys_Execute(); }
	inline gmThread::State ExecuteLibrary( const char* Filename ) { gmThread* pThread = LoadLibrary( Filename ); return pThread->Sys_Execute(); }

	bool CompileScript( const char* ScriptName, const char* LibraryName );

	gmType Bind( bool isClass, const char* Name, ...);

	inline gmVariableEx Call(const char* FuncName)																																																																{ return _CallFunc<0,int,int,int,int,int,int,int,int,int,int>( FuncName, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, gmVariable::s_null ); }
	template<typename P1> inline gmVariableEx Call(const char* FuncName,P1 Param1)																																																								{ return _CallFunc<1,P1,int,int,int,int,int,int,int,int,int>( FuncName, Param1, 0, 0, 0, 0, 0, 0, 0, 0, 0, gmVariable::s_null ); }
	template<typename P1, typename P2> inline gmVariableEx Call(const char* FuncName, P1 Param1, P2 Param2)																																																		{ return _CallFunc<2,P1,P2,int,int,int,int,int,int,int,int>( FuncName, Param1, Param2, 0, 0, 0, 0, 0, 0, 0, 0, gmVariable::s_null ); }
	template<typename P1, typename P2, typename P3> inline gmVariableEx Call(const char* FuncName, P1 Param1, P2 Param2, P3 Param3)																																												{ return _CallFunc<3,P1,P2,P3,int,int,int,int,int,int,int>( FuncName, Param1, Param2, Param3, 0, 0, 0, 0, 0, 0, 0, gmVariable::s_null ); }
	template<typename P1, typename P2, typename P3, typename P4> inline gmVariableEx Call(const char* FuncName, P1 Param1, P2 Param2, P3 Param3, P4 Param4)																																						{ return _CallFunc<4,P1,P2,P3,P4,int,int,int,int,int,int>( FuncName, Param1, Param2, Param3, Param4, 0, 0, 0, 0, 0, 0, gmVariable::s_null ); }
	template<typename P1, typename P2, typename P3, typename P4, typename P5> inline gmVariableEx Call(const char* FuncName, P1 Param1, P2 Param2, P3 Param3, P4 Param4, P5 Param5)																																{ return _CallFunc<5,P1,P2,P3,P4,P5,int,int,int,int,int>( FuncName, Param1, Param2, Param3, Param4, Param5, 0, 0, 0, 0, 0, gmVariable::s_null ); }
	template<typename P1, typename P2, typename P3, typename P4, typename P5, typename P6> inline gmVariableEx Call(const char* FuncName, P1 Param1, P2 Param2, P3 Param3, P4 Param4, P5 Param5, P6 Param6)																										{ return _CallFunc<6,P1,P2,P3,P4,P5,P6,int,int,int,int>( FuncName, Param1, Param2, Param3, Param4, Param5, Param6, 0, 0, 0, 0, gmVariable::s_null ); }
	template<typename P1, typename P2, typename P3, typename P4, typename P5, typename P6, typename P7> inline gmVariableEx Call(const char* FuncName, P1 Param1, P2 Param2, P3 Param3, P4 Param4, P5 Param5, P6 Param6, P7 Param7)																				{ return _CallFunc<7,P1,P2,P3,P4,P5,P6,P7,int,int,int>( FuncName, Param1, Param2, Param3, Param4, Param5, Param6, Param7, 0, 0, 0, gmVariable::s_null ); }
	template<typename P1, typename P2, typename P3, typename P4, typename P5, typename P6, typename P7, typename P8> inline gmVariableEx Call(const char* FuncName, P1 Param1, P2 Param2, P3 Param3, P4 Param4, P5 Param5, P6 Param6, P7 Param7, P8 Param8)														{ return _CallFunc<8,P1,P2,P3,P4,P5,P6,P7,P8,int,int>( FuncName, Param1, Param2, Param3, Param4, Param5, Param6, Param7, Param8, 0, 0, gmVariable::s_null ); }
	template<typename P1, typename P2, typename P3, typename P4, typename P5, typename P6, typename P7, typename P8, typename P9> inline gmVariableEx Call(const char* FuncName, P1 Param1, P2 Param2, P3 Param3, P4 Param4, P5 Param5, P6 Param6, P7 Param7, P8 Param8, P9 Param9)								{ return _CallFunc<9,P1,P2,P3,P4,P5,P6,P7,P8,P9,int>( FuncName, Param1, Param2, Param3, Param4, Param5, Param6, Param7, Param8, Param9, 0, gmVariable::s_null ); }
	template<typename P1, typename P2, typename P3, typename P4, typename P5, typename P6, typename P7, typename P8, typename P9, typename P10> inline gmVariableEx Call(const char* FuncName, P1 Param1, P2 Param2, P3 Param3, P4 Param4, P5 Param5, P6 Param6, P7 Param7, P8 Param8, P9 Param9, P10 Param10)	{ return _CallFunc<10,P1,P2,P3,P4,P5,P6,P7,P8,P9,P10>( FuncName, Param1, Param2, Param3, Param4, Param5, Param6, Param7, Param8, Param9, Param10, gmVariable::s_null ); }

	template<typename THIS> inline gmVariableEx Call(THIS This, const char* FuncName)																																																														{ return _CallFunc<0,int,int,int,int,int,int,int,int,int,int>( FuncName, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, CONVERT< GET<THIS>::Type, THIS >::ToGm(This,this, ( GET<THIS>::Type < GM_USER ) ? GET<THIS>::Type : GetTypeId<THIS>() ) ); }
	template<typename THIS, typename P1> inline gmVariableEx Call(THIS This, const char* FuncName,P1 Param1)																																																								{ return _CallFunc<1,P1,int,int,int,int,int,int,int,int,int>( FuncName, Param1, 0, 0, 0, 0, 0, 0, 0, 0, 0, CONVERT< GET<THIS>::Type, THIS >::ToGm(This,this, ( GET<THIS>::Type < GM_USER ) ? GET<THIS>::Type : GetTypeId<THIS>() ) ); }
	template<typename THIS, typename P1, typename P2> inline gmVariableEx Call(THIS This, const char* FuncName, P1 Param1, P2 Param2)																																																		{ return _CallFunc<2,P1,P2,int,int,int,int,int,int,int,int>( FuncName, Param1, Param2, 0, 0, 0, 0, 0, 0, 0, 0, CONVERT< GET<THIS>::Type, THIS >::ToGm(This,this, ( GET<THIS>::Type < GM_USER ) ? GET<THIS>::Type : GetTypeId<THIS>() ) ); }
	template<typename THIS, typename P1, typename P2, typename P3> inline gmVariableEx Call(THIS This, const char* FuncName, P1 Param1, P2 Param2, P3 Param3)																																												{ return _CallFunc<3,P1,P2,P3,int,int,int,int,int,int,int>( FuncName, Param1, Param2, Param3, 0, 0, 0, 0, 0, 0, 0, CONVERT< GET<THIS>::Type, THIS >::ToGm(This,this, ( GET<THIS>::Type < GM_USER ) ? GET<THIS>::Type : GetTypeId<THIS>() ) ); }
	template<typename THIS, typename P1, typename P2, typename P3, typename P4> inline gmVariableEx Call(THIS This, const char* FuncName, P1 Param1, P2 Param2, P3 Param3, P4 Param4)																																						{ return _CallFunc<4,P1,P2,P3,P4,int,int,int,int,int,int>( FuncName, Param1, Param2, Param3, Param4, 0, 0, 0, 0, 0, 0, CONVERT< GET<THIS>::Type, THIS >::ToGm(This,this, ( GET<THIS>::Type < GM_USER ) ? GET<THIS>::Type : GetTypeId<THIS>() ) ); }
	template<typename THIS, typename P1, typename P2, typename P3, typename P4, typename P5> inline gmVariableEx Call(THIS This, const char* FuncName, P1 Param1, P2 Param2, P3 Param3, P4 Param4, P5 Param5)																																{ return _CallFunc<5,P1,P2,P3,P4,P5,int,int,int,int,int>( FuncName, Param1, Param2, Param3, Param4, Param5, 0, 0, 0, 0, 0, CONVERT< GET<THIS>::Type, THIS >::ToGm(This,this, ( GET<THIS>::Type < GM_USER ) ? GET<THIS>::Type : GetTypeId<THIS>() ) ); }
	template<typename THIS, typename P1, typename P2, typename P3, typename P4, typename P5, typename P6> inline gmVariableEx Call(THIS This, const char* FuncName, P1 Param1, P2 Param2, P3 Param3, P4 Param4, P5 Param5, P6 Param6)																										{ return _CallFunc<6,P1,P2,P3,P4,P5,P6,int,int,int,int>( FuncName, Param1, Param2, Param3, Param4, Param5, Param6, 0, 0, 0, 0, CONVERT< GET<THIS>::Type, THIS >::ToGm(This,this, ( GET<THIS>::Type < GM_USER ) ? GET<THIS>::Type : GetTypeId<THIS>() ) ); }
	template<typename THIS, typename P1, typename P2, typename P3, typename P4, typename P5, typename P6, typename P7> inline gmVariableEx Call(THIS This, const char* FuncName, P1 Param1, P2 Param2, P3 Param3, P4 Param4, P5 Param5, P6 Param6, P7 Param7)																				{ return _CallFunc<7,P1,P2,P3,P4,P5,P6,P7,int,int,int>( FuncName, Param1, Param2, Param3, Param4, Param5, Param6, Param7, 0, 0, 0, CONVERT< GET<THIS>::Type, THIS >::ToGm(This,this, ( GET<THIS>::Type < GM_USER ) ? GET<THIS>::Type : GetTypeId<THIS>() ) ); }
	template<typename THIS, typename P1, typename P2, typename P3, typename P4, typename P5, typename P6, typename P7, typename P8> inline gmVariableEx Call(THIS This, const char* FuncName, P1 Param1, P2 Param2, P3 Param3, P4 Param4, P5 Param5, P6 Param6, P7 Param7, P8 Param8)														{ return _CallFunc<8,P1,P2,P3,P4,P5,P6,P7,P8,int,int>( FuncName, Param1, Param2, Param3, Param4, Param5, Param6, Param7, Param8, 0, 0, CONVERT< GET<THIS>::Type, THIS >::ToGm(This,this, ( GET<THIS>::Type < GM_USER ) ? GET<THIS>::Type : GetTypeId<THIS>() ) ); }
	template<typename THIS, typename P1, typename P2, typename P3, typename P4, typename P5, typename P6, typename P7, typename P8, typename P9> inline gmVariableEx Call(THIS This, const char* FuncName, P1 Param1, P2 Param2, P3 Param3, P4 Param4, P5 Param5, P6 Param6, P7 Param7, P8 Param8, P9 Param9)								{ return _CallFunc<9,P1,P2,P3,P4,P5,P6,P7,P8,P9,int>( FuncName, Param1, Param2, Param3, Param4, Param5, Param6, Param7, Param8, Param9, 0, CONVERT< GET<THIS>::Type, THIS >::ToGm(This,this, ( GET<THIS>::Type < GM_USER ) ? GET<THIS>::Type : GetTypeId<THIS>() ) ); }
	template<typename THIS, typename P1, typename P2, typename P3, typename P4, typename P5, typename P6, typename P7, typename P8, typename P9, typename P10> inline gmVariableEx Call(THIS This, const char* FuncName, P1 Param1, P2 Param2, P3 Param3, P4 Param4, P5 Param5, P6 Param6, P7 Param7, P8 Param8, P9 Param9, P10 Param10)	{ return _CallFunc<10,P1,P2,P3,P4,P5,P6,P7,P8,P9,P10>( FuncName, Param1, Param2, Param3, Param4, Param5, Param6, Param7, Param8, Param9, Param10, CONVERT< GET<THIS>::Type, THIS >::ToGm(This,this, ( GET<THIS>::Type < GM_USER ) ? GET<THIS>::Type : GetTypeId<THIS>() ) ); }

	static std::map<const gmType,gmClassDef*> ClassList;
	static std::multimap<const int,gmProxy*> CppFuncList; 

	static std::multimap<const int,gmThread*> _ThreadList_AtCreate;
	static std::multimap<const int,gmThread*> _ThreadList_AtDestroy;

	static int gmThreadEx(gmThread* pthread);
	static bool gmMachineEx::OnCreateThread( const void* pContext );
	static bool gmMachineEx::OnDestroyThread( const void* pContext );
	static bool MachineCallback( gmMachine*, gmMachineCommand Command, const void* pContext);

	static int GM_CDECL FunctionDispatcher( gmThread* );
	static int GM_CDECL MethodeDispatcher( gmThread* );
	static int GM_CDECL OperatorDispatcher( gmThread* );

	static void GM_CDECL OpGetDot(gmThread*, gmVariable*);
	static void GM_CDECL OpSetDot(gmThread*, gmVariable*);

	static unsigned int Hash( const char *Text );
	static gmOperator CppOpToGm( const char*, bool );

private:
	char _Script[65536];
	unsigned int _CurrentHash;
	unsigned int _GlobalFuncIndex;
	gmFunctionObject* _CurrentFunc;
	std::map<const unsigned int,gmFunctionObject*> _GmFuncList;

	std::vector<int*> _IndexList;

	int AT_CREATE, AT_DESTROY, AT_TIME;

	gmThread* _BeginFuncCall( const char* FuncName, const gmVariable& This );
	gmVariableEx _EndFuncCall( gmThread* pThread, int ParamNumber );
	template<int NumParam, typename P1, typename P2, typename P3, typename P4, typename P5, typename P6, typename P7, typename P8, typename P9, typename P10> inline gmVariableEx _CallFunc( const char* FuncName, P1 Param1, P2 Param2, P3 Param3, P4 Param4, P5 Param5, P6 Param6, P7 Param7, P8 Param8, P9 Param9, P10 Param10, const gmVariable& This );
};

template<int NumParam, typename P1, typename P2, typename P3, typename P4, typename P5, typename P6, typename P7, typename P8, typename P9, typename P10> inline gmVariableEx gmMachineEx::_CallFunc( const char* FuncName, P1 Param1, P2 Param2, P3 Param3, P4 Param4, P5 Param5, P6 Param6, P7 Param7, P8 Param8, P9 Param9, P10 Param10, const gmVariable& This )
{
	gmThread* pThread = _BeginFuncCall( FuncName, This );

	if( pThread )
	{
		IF< ( NumParam > 0 ) >::Push<P1>( Param1, pThread );
		IF< ( NumParam > 1 ) >::Push<P2>( Param2, pThread );
		IF< ( NumParam > 2 ) >::Push<P3>( Param3, pThread );
		IF< ( NumParam > 3 ) >::Push<P4>( Param4, pThread );
		IF< ( NumParam > 4 ) >::Push<P5>( Param5, pThread );
		IF< ( NumParam > 5 ) >::Push<P6>( Param6, pThread );
		IF< ( NumParam > 6 ) >::Push<P7>( Param7, pThread );
		IF< ( NumParam > 7 ) >::Push<P8>( Param8, pThread );
		IF< ( NumParam > 8 ) >::Push<P9>( Param9, pThread );
		IF< ( NumParam > 9 ) >::Push<P10>( Param10, pThread );
	}

	return _EndFuncCall( pThread, NumParam );
}

template<class _CLASS_> gmType GetTypeId()
{
	const type_info* pThisInfo = &typeid( ADJUST<_CLASS_>::TYPE );

	std::map<const gmType,gmClassDef*>::iterator i = gmMachineEx::ClassList.begin();

	while( i != gmMachineEx::ClassList.end() )
	{
		if( *(i->second->pTypeInfo) == *pThisInfo )
			return i->first;
		else
			++i;
	}

	return GM_UNKNOWN;
};

gmType GetTypeId( type_info* pInfo );


#pragma warning ( default: 4244 )

#endif