#ifndef GM_VARIABLE_PROXY_H
#define GM_VARIABLE_PROXY_H

template<class CLASS, typename TYPE, bool AccessType> class gmVarProxy : public gmProxy
{
public:
	gmVarProxy( TYPE CLASS::*pVar )
	{
		this->pVar = pVar;
		GmType = ( GET<TYPE>::Type >= GM_NULL && GET<TYPE>::Type < GM_USER ) ? GET<TYPE>::Type : GetTypeId<TYPE>();
	}

	gmVarProxy( TYPE CLASS::*pVar, TYPE (CLASS::*pGet)(), void (CLASS::*pSet)(TYPE) )
	{
		this->pVar = pVar;
		this->pGetSet.pGet = pGet;
		this->pGetSet.pSet = pSet;
		GmType = ( GET<TYPE>::Type >= GM_NULL && GET<TYPE>::Type < GM_USER ) ? GET<TYPE>::Type : GetTypeId<TYPE>();
	}

	void operator()(void* pObj, gmVariable& Var) const
	{
		IF< AccessType == DIRECT_ACCESS >::Set<CLASS,TYPE>( (CLASS*)pObj, pVar, pGetSet.pSet, Var );
	}

	gmVariable operator()(gmThread* pThread, void* pObj ) const
	{
		return IF< AccessType == DIRECT_ACCESS >::Get<CLASS,TYPE>( (CLASS*)pObj, pVar, pGetSet.pGet, pThread->GetMachine(), GmType );
	}

	int GetProxyType() const { return VARIABLE_TYPE; }
	void UpdateTypes() { if( GmType == GM_UNKNOWN ) GmType = GetTypeId<TYPE>(); }

	int CompareTypes( gmType Type, int ) const
	{
		if( Type == GmType ) return true;

		switch( Type )
		{
		case GM_NULL:
			return false;

		case GM_INT:
			{
				if( GmType == GM_FLOAT || GmType == GM_STRING || GmType == GM_C_STRING || GmType == GM_CONST_C_STRING )
					return true;

			} break;

		case GM_FLOAT:
			{
				if( GmType == GM_INT || GmType == GM_STRING || GmType == GM_C_STRING || GmType == GM_CONST_C_STRING )
					return true;

			} break;

		case GM_STRING:
			{
				if( GmType == GM_INT || GmType == GM_FLOAT || GmType == GM_C_STRING || GmType == GM_CONST_C_STRING )
					return true;

			} break;

	
		case GM_CONST_C_STRING:
		case GM_C_STRING:
			{
				if( GmType == GM_STRING || GmType == GM_INT || GmType == GM_FLOAT )
					return true;		

			} break;
				
		case GM_TABLE:
		case GM_FUNCTION:
			{
				if( GmType == GM_ANY_PTR || GmType == GM_CONST_ANY_PTR )
					return true;

			} break;
			
		default:
			{
				if( Type % 2 ) //Type est non const ptr
				{
					if( GmType == Type+1 || GmType == GM_ANY_PTR || GmType == GM_CONST_ANY_PTR )
						return true;
				}
				else //Type est const ptr
				{
					if( GmType == -(Type-1) || GmType == GM_CONST_ANY_PTR )
						return true;
					else if( GmType == GM_ANY_PTR || GmType == Type-1 )
						return false;
				}

				std::map<const gmType,gmClassDef*>::const_iterator i = gmMachineEx::ClassList.find( Type );
				if( i != gmMachineEx::ClassList.end() )
				{
					std::map<const gmType,gmClassDef*>::const_iterator j = gmMachineEx::ClassList.find( GmType );

					if( j != gmMachineEx::ClassList.end() )
					{
						if( i->second->Base == j->second )
							return true;
					}

				}

			}	
		}
		
		return false;
	};

private:
	union
	{
		TYPE CLASS::*pVar;
		struct
		{
			TYPE (CLASS::*pGet)();
			void (CLASS::*pSet)(TYPE);
		} pGetSet;
	};

	gmType GmType;
};


template<class CLASS, typename TYPE> inline gmProxy* CreateVariableProxy( TYPE CLASS::*pVar )
{
	return new gmVarProxy<CLASS,TYPE,DIRECT_ACCESS>( pVar );
}

template<class CLASS, typename TYPE> inline gmProxy* CreateVariableProxy( TYPE CLASS::*pVar, TYPE(CLASS::*pGet)(), void(CLASS::*pSet)(TYPE) )
{
	return new gmVarProxy<CLASS,TYPE,GET_SET_ACCESS>( pVar, pGet, pSet );
}

#endif